--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;

package chargen_pkg is

  type fontcfg_typ is record
    width  : positive;
    height : positive;
    origin : natural;
    count  : positive;
  end record;

  constant fontcfg : fontcfg_typ := (
    width  => 8,
    height => 8,
    origin => 32,
    count  => 95);

  constant charnum_high : positive := videotiming.hvisible / fontcfg.width;

  subtype linenum_typ is natural range 0 to fontcfg.height - 1;
  subtype colnum_typ is natural range 0 to fontcfg.width - 1;
  subtype charnum_typ is natural range 0 to charnum_high;

  component character_generator is
    port (chr : in character; linenum : in linenum_typ; colnum : in colnum_typ; pix : out std_ulogic);
  end component;

  component string_generator is
    port(clk, ena, next_line, rowstart, colstart, last_char, last_string : in std_ulogic; linenum : out linenum_typ; colnum : out colnum_typ; charnum : out charnum_typ; next_string, strobe : out std_ulogic);
  end component;

  component row_ff is
    port (clk, ena, next_line, rowstart, next_string, last_string, colstart : std_ulogic; subcolstart, rowstrobe : out std_ulogic);
  end component;

  function fontrow(c : character; r : natural) return std_ulogic_vector;

end chargen_pkg;

library ieee;
use ieee.std_logic_1164.all;

entity row_ff is
  port (clk, ena, next_line, rowstart, next_string, last_string, colstart : in std_ulogic; subcolstart, rowstrobe : out std_ulogic);
end row_ff;

architecture arch_row_ff of row_ff is
  signal ff : std_ulogic;
begin

  process (clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = next_line then
          if '1' = (last_string and next_string) then
            ff <= '0';
          else
            if '1' = rowstart then
              ff <= '1';
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  subcolstart <= colstart and (rowstart or ff);

  rowstrobe <= rowstart or ff;

end arch_row_ff;

library ieee;
use ieee.std_logic_1164.all;
use work.chargen_pkg.all;

entity character_generator is
  port (chr : in character; linenum : in linenum_typ; colnum : in colnum_typ; pix : out std_ulogic);
end character_generator;

architecture arch_character_generator of character_generator is
begin

  pix <= fontrow(chr, linenum)(colnum);

end arch_character_generator;

library ieee;
use ieee.std_logic_1164.all;
use work.chargen_pkg.all;

entity string_generator is
  port(clk, ena, next_line, rowstart, colstart, last_char, last_string : in std_ulogic; linenum : out linenum_typ; colnum : out colnum_typ; charnum : out charnum_typ; next_string, strobe : out std_ulogic);
end string_generator;

architecture arch_string_generator of string_generator is

  signal linecnt      : linenum_typ;
  signal columncnt    : colnum_typ;
  signal charcnt      : charnum_typ;
  signal rowff, colff : std_ulogic;
  signal last_line    : boolean;

begin

  process (clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = next_line then

          if '1' = rowstart then
            linecnt <= 1;
          else
            if last_line then
              linecnt <= 0;
            else
              linecnt <= linecnt + 1;
            end if;
          end if;

        end if;
      end if;
    end if;
  end process;

  last_line <= (fontcfg.height - 1 <= linecnt);

  process (clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = next_line then

          if ('1' = last_string) and last_line then
            rowff <= '0';
          else
            if '1' = rowstart then
              rowff <= '1';
            end if;
          end if;

        end if;
      end if;
    end if;
  end process;

  next_string <= '1' when last_line else '0';

  process (clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = colstart then
          columncnt <= 1;
          charcnt   <= 0;
          colff     <= '1';
        else
          if fontcfg.width - 1 <= columncnt then
            columncnt <= 0;
            if charnum_typ'high > charcnt then
              charcnt <= charcnt + 1;
            else
              charcnt <= 0;
            end if;

            if '1' = last_char then
              colff <= '0';
            end if;
          else
            columncnt <= columncnt + 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  linenum <= 0 when '1' = rowstart else linecnt;

  colnum <= 0 when '1' = colstart else columncnt;

  charnum <= 0 when '1' = colstart else charcnt;

  strobe <= (rowstart or rowff) and (colstart or colff);

end arch_string_generator;

package body chargen_pkg is

  subtype ftrow_typ is std_ulogic_vector(0 to fontcfg.width - 1);

  type font_typ is array (natural range <>) of ftrow_typ;

  constant fontglyph_use : font_typ(fontcfg.origin * fontcfg.height to (fontcfg.origin + fontcfg.count) * fontcfg.height - 1) := (
--0x08,0x08,0x20,0x5F,
    X"00", X"00", X"00", X"00", X"00", X"00", X"00", X"00",  -- <Space>
    X"18", X"3C", X"3C", X"18", X"18", X"00", X"18", X"00",  -- !
    X"66", X"66", X"24", X"00", X"00", X"00", X"00", X"00",  -- "
    X"6C", X"6C", X"FE", X"6C", X"FE", X"6C", X"6C", X"00",  -- #
    X"18", X"3E", X"60", X"3C", X"06", X"7C", X"18", X"00",  -- $
    X"00", X"C6", X"CC", X"18", X"30", X"66", X"C6", X"00",  -- %
    X"38", X"6C", X"38", X"76", X"DC", X"CC", X"76", X"00",  -- &
    X"18", X"18", X"30", X"00", X"00", X"00", X"00", X"00",  -- '
    X"0C", X"18", X"30", X"30", X"30", X"18", X"0C", X"00",  -- (
    X"30", X"18", X"0C", X"0C", X"0C", X"18", X"30", X"00",  -- )
    X"00", X"66", X"3C", X"FF", X"3C", X"66", X"00", X"00",  -- *
    X"00", X"18", X"18", X"7E", X"18", X"18", X"00", X"00",  -- +
    X"00", X"00", X"00", X"00", X"00", X"18", X"18", X"30",  -- ,
    X"00", X"00", X"00", X"7E", X"00", X"00", X"00", X"00",  -- -
    X"00", X"00", X"00", X"00", X"00", X"18", X"18", X"00",  -- .
    X"06", X"0C", X"18", X"30", X"60", X"C0", X"80", X"00",  -- /
    X"7C", X"C6", X"CE", X"D6", X"E6", X"C6", X"7C", X"00",  -- 0
    X"18", X"38", X"18", X"18", X"18", X"18", X"7E", X"00",  -- 1
    X"7C", X"C6", X"06", X"1C", X"30", X"66", X"FE", X"00",  -- 2
    X"7C", X"C6", X"06", X"3C", X"06", X"C6", X"7C", X"00",  -- 3
    X"1C", X"3C", X"6C", X"CC", X"FE", X"0C", X"1E", X"00",  -- 4
    X"FE", X"C0", X"C0", X"FC", X"06", X"C6", X"7C", X"00",  -- 5
    X"38", X"60", X"C0", X"FC", X"C6", X"C6", X"7C", X"00",  -- 6
    X"FE", X"C6", X"0C", X"18", X"30", X"30", X"30", X"00",  -- 7
    X"7C", X"C6", X"C6", X"7C", X"C6", X"C6", X"7C", X"00",  -- 8
    X"7C", X"C6", X"C6", X"7E", X"06", X"0C", X"78", X"00",  -- 9
    X"00", X"18", X"18", X"00", X"00", X"18", X"18", X"00",  -- :
    X"00", X"18", X"18", X"00", X"00", X"18", X"18", X"30",  -- ;
    X"06", X"0C", X"18", X"30", X"18", X"0C", X"06", X"00",  -- <
    X"00", X"00", X"7E", X"00", X"00", X"7E", X"00", X"00",  -- =
    X"60", X"30", X"18", X"0C", X"18", X"30", X"60", X"00",  -- >
    X"7C", X"C6", X"0C", X"18", X"18", X"00", X"18", X"00",  -- ?
    X"7C", X"C6", X"DE", X"DE", X"DE", X"C0", X"78", X"00",  -- @
    X"38", X"6C", X"C6", X"FE", X"C6", X"C6", X"C6", X"00",  -- A
    X"FC", X"66", X"66", X"7C", X"66", X"66", X"FC", X"00",  -- B
    X"3C", X"66", X"C0", X"C0", X"C0", X"66", X"3C", X"00",  -- C
    X"F8", X"6C", X"66", X"66", X"66", X"6C", X"F8", X"00",  -- D
    X"FE", X"62", X"68", X"78", X"68", X"62", X"FE", X"00",  -- E
    X"FE", X"62", X"68", X"78", X"68", X"60", X"F0", X"00",  -- F
    X"3C", X"66", X"C0", X"C0", X"CE", X"66", X"3A", X"00",  -- G
    X"C6", X"C6", X"C6", X"FE", X"C6", X"C6", X"C6", X"00",  -- H
    X"3C", X"18", X"18", X"18", X"18", X"18", X"3C", X"00",  -- I
    X"1E", X"0C", X"0C", X"0C", X"CC", X"CC", X"78", X"00",  -- J
    X"E6", X"66", X"6C", X"78", X"6C", X"66", X"E6", X"00",  -- K
    X"F0", X"60", X"60", X"60", X"62", X"66", X"FE", X"00",  -- L
    X"C6", X"EE", X"FE", X"FE", X"D6", X"C6", X"C6", X"00",  -- M
    X"C6", X"E6", X"F6", X"DE", X"CE", X"C6", X"C6", X"00",  -- N
    X"7C", X"C6", X"C6", X"C6", X"C6", X"C6", X"7C", X"00",  -- O
    X"FC", X"66", X"66", X"7C", X"60", X"60", X"F0", X"00",  -- P
    X"7C", X"C6", X"C6", X"C6", X"C6", X"CE", X"7C", X"0E",  -- Q
    X"FC", X"66", X"66", X"7C", X"6C", X"66", X"E6", X"00",  -- R
    X"7C", X"C6", X"60", X"38", X"0C", X"C6", X"7C", X"00",  -- S
    X"7E", X"7E", X"5A", X"18", X"18", X"18", X"3C", X"00",  -- T
    X"C6", X"C6", X"C6", X"C6", X"C6", X"C6", X"7C", X"00",  -- U
    X"C6", X"C6", X"C6", X"C6", X"C6", X"6C", X"38", X"00",  -- V
    X"C6", X"C6", X"C6", X"D6", X"D6", X"FE", X"6C", X"00",  -- W
    X"C6", X"C6", X"6C", X"38", X"6C", X"C6", X"C6", X"00",  -- X
    X"66", X"66", X"66", X"3C", X"18", X"18", X"3C", X"00",  -- Y
    X"FE", X"C6", X"8C", X"18", X"32", X"66", X"FE", X"00",  -- Z
    X"3C", X"30", X"30", X"30", X"30", X"30", X"3C", X"00",  -- [
    X"C0", X"60", X"30", X"18", X"0C", X"06", X"02", X"00",  -- <Backslash>
    X"3C", X"0C", X"0C", X"0C", X"0C", X"0C", X"3C", X"00",  -- ]
    X"10", X"38", X"6C", X"C6", X"00", X"00", X"00", X"00",  -- ^
    X"00", X"00", X"00", X"00", X"00", X"00", X"00", X"FF",  -- _
    X"30", X"18", X"0C", X"00", X"00", X"00", X"00", X"00",  -- '
    X"00", X"00", X"78", X"0C", X"7C", X"CC", X"76", X"00",  -- a
    X"E0", X"60", X"7C", X"66", X"66", X"66", X"DC", X"00",  -- b
    X"00", X"00", X"7C", X"C6", X"C0", X"C6", X"7C", X"00",  -- c
    X"1C", X"0C", X"7C", X"CC", X"CC", X"CC", X"76", X"00",  -- d
    X"00", X"00", X"7C", X"C6", X"FE", X"C0", X"7C", X"00",  -- e
    X"3C", X"66", X"60", X"F8", X"60", X"60", X"F0", X"00",  -- f
    X"00", X"00", X"76", X"CC", X"CC", X"7C", X"0C", X"F8",  -- g
    X"E0", X"60", X"6C", X"76", X"66", X"66", X"E6", X"00",  -- h
    X"18", X"00", X"38", X"18", X"18", X"18", X"3C", X"00",  -- i
    X"06", X"00", X"06", X"06", X"06", X"66", X"66", X"3C",  -- j
    X"E0", X"60", X"66", X"6C", X"78", X"6C", X"E6", X"00",  -- k
    X"38", X"18", X"18", X"18", X"18", X"18", X"3C", X"00",  -- l
    X"00", X"00", X"EC", X"FE", X"D6", X"D6", X"D6", X"00",  -- m
    X"00", X"00", X"DC", X"66", X"66", X"66", X"66", X"00",  -- n
    X"00", X"00", X"7C", X"C6", X"C6", X"C6", X"7C", X"00",  -- o
    X"00", X"00", X"DC", X"66", X"66", X"7C", X"60", X"F0",  -- p
    X"00", X"00", X"76", X"CC", X"CC", X"7C", X"0C", X"1E",  -- q
    X"00", X"00", X"DC", X"76", X"60", X"60", X"F0", X"00",  -- r
    X"00", X"00", X"7E", X"C0", X"7C", X"06", X"FC", X"00",  -- s
    X"30", X"30", X"FC", X"30", X"30", X"36", X"1C", X"00",  -- t
    X"00", X"00", X"CC", X"CC", X"CC", X"CC", X"76", X"00",  -- u
    X"00", X"00", X"C6", X"C6", X"C6", X"6C", X"38", X"00",  -- v
    X"00", X"00", X"C6", X"D6", X"D6", X"FE", X"6C", X"00",  -- w
    X"00", X"00", X"C6", X"6C", X"38", X"6C", X"C6", X"00",  -- x
    X"00", X"00", X"C6", X"C6", X"C6", X"7E", X"06", X"FC",  -- y
    X"00", X"00", X"7E", X"4C", X"18", X"32", X"7E", X"00",  -- z
    X"0E", X"18", X"18", X"70", X"18", X"18", X"0E", X"00",  -- {
    X"18", X"18", X"18", X"18", X"18", X"18", X"18", X"00",  -- |
    X"70", X"18", X"18", X"0E", X"18", X"18", X"70", X"00",  -- }
    X"76", X"DC", X"00", X"00", X"00", X"00", X"00", X"00"   -- ~
    );

  function fillfont(v : ftrow_typ; n : natural) return font_typ is
    variable t : font_typ(0 to n - 1);
  begin
    t := (others => v);
    return t;
  end;

  function fillv(a : std_ulogic; n : natural) return std_ulogic_vector is
    variable t : std_ulogic_vector(0 to n - 1);
  begin
    t := (others => a);
    return t;
  end;

  constant fontglyph : font_typ(0 to (fontcfg.origin + fontcfg.count) * fontcfg.height - 1) :=
    fillfont(fillv('0', fontcfg.width), fontcfg.origin * fontcfg.height) & fontglyph_use;

  function fontrow(c : character; r : natural) return std_ulogic_vector is
    variable fgr : ftrow_typ;
  begin
    fgr := fontglyph(character'pos(c) * fontcfg.height + r);
    return fgr;
  end;

  function fontrow1(c : character; r : natural) return std_ulogic_vector is
    variable fa : font_typ(fontcfg.origin * fontcfg.height to (fontcfg.origin + fontcfg.count) * fontcfg.height - 1);
  begin
    for i in fontcfg.origin * fontcfg.height to (fontcfg.origin + fontcfg.count) * fontcfg.height - 1 loop
      fa(i) := fontglyph(i);
    end loop;
    return fa((character'pos(c)) * fontcfg.height + r);
  end;

end chargen_pkg;
