--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;
use work.fpotyp_pkg.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;
use work.chargen_pkg.all;

package gobjs_pkg is

  type color_typ is record
    r, g, b : std_ulogic;
  end record;

  component colstrobe is
    port (clk, ena : in std_ulogic; col, colbegin, colend : in col_typ; strobe : out std_ulogic);
  end component;

  component rowstrobe is
    port (clk, ena, next_line : in std_ulogic; row, rowbegin, rowend : in row_typ; strobe : out std_ulogic);
  end component;

  component trgstring is
    port(clk, ena, vactive, ready, fill, run : in std_ulogic; charnum : in charnum_typ; last_char : out std_ulogic; chr : out character; trigger_color : out color_typ);
  end component;

  component modestring is
    port(trg_mod : in trg_mod_typ; charnum : in charnum_typ; last_char : out std_ulogic; chr : out character; mode_color : out color_typ);
  end component;

  component gridline is
    port (clk, ena : in std_ulogic; pix : out std_ulogic; col : in col_typ; line : in std_ulogic_vector(3 downto 0));
  end component;

  component statusstring is
    port (charnum : in charnum_typ; capture_ctrl : capture_ctrl_rec; display_ctrl : in display_ctrl_rec; cur : in ramadr_typ; last_char : out std_ulogic; chr : out character);
  end component;

  component controlcursor is
    port (clk, ena : in std_ulogic; pix : out std_ulogic; x, col : in col_typ; y, row : in row_typ);
  end component;

end gobjs_pkg;

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;

entity colstrobe is
  port (clk, ena : in std_ulogic; col, colbegin, colend : in col_typ; strobe : out std_ulogic);
end colstrobe;

architecture arch_colstrobe of colstrobe is

  signal ff, bc, ec : std_ulogic;

begin

  bc <= '1' when colbegin = col else '0';

  ec <= '1' when colend = col else '0';

  process(clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = ec then
          ff <= '0';
        else
          if '1' = bc then
            ff <= '1';
          end if;
        end if;
      end if;
    end if;
  end process;

  strobe <= bc or ff;

end arch_colstrobe;

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;

entity rowstrobe is
  port (clk, ena, next_line : in std_ulogic; row, rowbegin, rowend : in row_typ; strobe : out std_ulogic);
end rowstrobe;

architecture arch_rowstrobe of rowstrobe is

  signal ff, br, er : std_ulogic;

begin

  br <= '1' when rowbegin = row else '0';

  er <= '1' when rowend = row else '0';

  process(clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = next_line then
          if '1' = er then
            ff <= '0';
          else
            if '1' = br then
              ff <= '1';
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  strobe <= br or ff;

end arch_rowstrobe;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.videomode_pkg.all;

entity gridline is
  port (clk, ena : in std_ulogic; pix : out std_ulogic; col : in col_typ; line : in std_ulogic_vector(3 downto 0));
end gridline;

architecture arch_gridline of gridline is
begin

  pix <= '1' when (((("00" = line(1 downto 0)) or ("11" = line(1 downto 0))) and ("00000" = col(4 downto 0))) or ("1111" = line(3 downto 0))) else '0';

end arch_gridline;

library ieee;
use ieee.std_logic_1164.all;
use work.fpotyp_pkg.all;
use work.chargen_pkg.all;
use work.gobjs_pkg.all;

entity trgstring is
  port(clk, ena, vactive, ready, fill, run : in std_ulogic; charnum : in charnum_typ; last_char : out std_ulogic; chr : out character; trigger_color : out color_typ);
end trgstring;

architecture arch_trgstring of trgstring is

  signal mready, mfill, mrun, rr, gg, bb : std_ulogic;

  type trg_string_typ is array (0 to 2) of character;

  signal trg_string : trg_string_typ;

  function lim3(x : natural) return natural is
    variable t : natural;
  begin
    t := x rem 4;
    if t >= 3 then
      t := 0;
    end if;
    return t;
  end;

begin

  process(clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' /= vactive then
          mready <= ready;
          mfill  <= fill;
          mrun   <= run;
        end if;
      end if;
    end if;
  end process;

  process(mready, mfill, mrun)
  begin
    if '1' = mrun then
      trg_string <= "RUN";
      rr         <= '0';
      gg         <= '1';
      bb         <= '0';
    else
      if '1' = mfill then
        trg_string <= "FIL";
        rr         <= '1';
        gg         <= '0';
        bb         <= '1';
      else
        if '1' = mready then
          trg_string <= "RDY";
          rr         <= '1';
          gg         <= '1';
          bb         <= '0';
        else
          trg_string <= "STP";
          rr         <= '1';
          gg         <= '0';
          bb         <= '0';
        end if;
      end if;
    end if;
  end process;

  trigger_color.r <= rr;
  trigger_color.g <= gg;
  trigger_color.b <= bb;

  chr <= trg_string(lim3(charnum));

  last_char <= '1' when (trg_string'length - 1) = lim3(charnum) else '0';

end arch_trgstring;

library ieee;
use ieee.std_logic_1164.all;
use work.fpotyp_pkg.all;
use work.chargen_pkg.all;
use work.gobjs_pkg.all;

entity modestring is
  port(trg_mod : in trg_mod_typ; charnum : in charnum_typ; last_char : out std_ulogic; chr : out character; mode_color : out color_typ);
end modestring;

architecture arch_modestring of modestring is

  type mode_string_typ is array (0 to 3) of character;

  signal mode_string : mode_string_typ;

begin

  mode_string <= "NORM" when (norm = trg_mod) else "SNGL";

  mode_color.r <= '0' when (norm = trg_mod) else '1';
  mode_color.g <= '1';
  mode_color.b <= '0';

  chr <= mode_string(charnum rem 4);

  last_char <= '1' when (mode_string'length - 1) = (charnum rem 4) else '0';

end arch_modestring;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.videomode_pkg.all;
use work.gobjs_pkg.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;
use work.chargen_pkg.all;

entity statusstring is
  port (charnum : in charnum_typ; capture_ctrl : capture_ctrl_rec; display_ctrl : in display_ctrl_rec; cur : in ramadr_typ; last_char : out std_ulogic; chr : out character);
end statusstring;

architecture arch_statusstring of statusstring is

  type char_array_typ is array (natural range <>) of character;
  type char_map_typ is array (0 to 15) of character;

  constant char_map : char_map_typ := ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');

  function hex2char(x : std_ulogic_vector) return char_array_typ is
    variable t : char_array_typ(0 to (x'length + 3)/4-1);  -- := "";
    variable y : std_ulogic_vector((x'length + 3)/4*4 - 1 downto 0) := (others => '0');
  begin
    y(x'high downto x'low) := x;
    for i in t'range loop
      t(t'high - i) := char_map(to_integer(unsigned(y(4*i+3 downto 4*i))));
    end loop;
    return t;
  end;

  function shex2char(x : std_ulogic_vector) return char_array_typ is
    variable t   : char_array_typ(0 to (x'length + 3)/4-1);  -- := "";
    variable y   : std_ulogic_vector((x'length + 3)/4*4 - 1 downto 0) := (others => '0');
    variable sig : character;
  begin
    if x(x'left) = '1' then
      y(x'high downto x'low) := std_ulogic_vector(to_unsigned(0, x'length) - unsigned(x));
      sig                    := '-';
    else
      y(x'high downto x'low) := x;
      sig                    := '+';
    end if;
    for i in t'range loop
      t(t'high - i) := char_map(to_integer(unsigned(y(4*i+3 downto 4*i))));
    end loop;
    return sig & t;
  end;

  constant status : char_array_typ(0 to 60) := "PIX/CLK=P/1   | DRIFT=SXXXX | DISP=XXX | DEPTH=XXXX | CUR=???";

  signal pbc, sig, dig : character;
  signal driftv        : std_ulogic_vector(15 downto 0);
  signal nibble        : std_ulogic_vector(3 downto 0);
  signal dispv         : std_ulogic_vector(11 downto 0);
  signal depth         : std_ulogic_vector(15 downto 0);
  signal curv          : std_ulogic_vector(11 downto 0);

  function lim60(x : natural) return natural is
    variable t : natural;
  begin
    t := x rem 64;
    if t > 60 then
      t := 0;
    end if;
    return t;
  end;

begin

  pbc <= '8' when 3 = display_ctrl.scale else '4' when 2 = display_ctrl.scale else '2' when 1 = display_ctrl.scale else '1';

  sig <= '+' when 0 < capture_ctrl.drift else '-' when 0 > capture_ctrl.drift else ' ';

  driftv <= std_ulogic_vector(to_signed(- capture_ctrl.drift, 16)) when capture_ctrl.drift < 0 else std_ulogic_vector(to_signed(capture_ctrl.drift, 16));

  dispv <= std_ulogic_vector(to_unsigned(display_ctrl.disp, 12));

  depth <= std_ulogic_vector(to_unsigned(capture_ctrl.depth, 16));

  curv <= std_ulogic_vector(to_unsigned(cur, 12));

  process(charnum, pbc, sig, dig, driftv, dispv, curv, depth)
  begin
    case charnum is
      when 8 =>
        chr    <= pbc;
        nibble <= (others => '0');
      when 22 =>
        chr    <= sig;
        nibble <= (others => '0');
      when 23 =>
        chr    <= dig;
        nibble <= driftv(15 downto 12);
      when 24 =>
        chr    <= dig;
        nibble <= driftv(11 downto 8);
      when 25 =>
        chr    <= dig;
        nibble <= driftv(7 downto 4);
      when 26 =>
        chr    <= dig;
        nibble <= driftv(3 downto 0);
      when 35 =>
        chr    <= dig;
        nibble <= dispv(11 downto 8);
      when 36 =>
        chr    <= dig;
        nibble <= dispv(7 downto 4);
      when 37 =>
        chr    <= dig;
        nibble <= dispv(3 downto 0);
      when 47 =>
        chr    <= dig;
        nibble <= depth(15 downto 12);
      when 48 =>
        chr    <= dig;
        nibble <= depth(11 downto 8);
      when 49 =>
        chr    <= dig;
        nibble <= depth(7 downto 4);
      when 50 =>
        chr    <= dig;
        nibble <= depth(3 downto 0);
      when 58 =>
        chr    <= dig;
        nibble <= curv(11 downto 8);
      when 59 =>
        chr    <= dig;
        nibble <= curv(7 downto 4);
      when 60 =>
        chr    <= dig;
        nibble <= curv(3 downto 0);

      when others =>
        chr    <= status(lim60(charnum));
        nibble <= (others => '0');
    end case;
  end process;

  dig <= char_map(to_integer(unsigned(nibble)));

  last_char <= '1' when (status'length - 1) = charnum else '0';

end arch_statusstring;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.videomode_pkg.all;

entity controlcursor is
  port (clk, ena : in std_ulogic; pix : out std_ulogic; x, col : in col_typ; y, row : in row_typ);
end controlcursor;

architecture arch_controlcursor of controlcursor is

  signal sx : signed(wh(videotiming.hvisible) + 1 downto 0);
  signal sy : signed(wh(videotiming.vvisible) + 1 downto 0);

begin

  sx <= signed('0' & col) - (('0' & signed(x)) - 8);
  sy <= signed('0' & row) - (('0' & signed(y)) - 8);

  pix <= '1' when ((0 <= sx) and (16 > sx) and (0 <= sy) and (16 > sy) and not ((sx(2 downto 0) = "000") or (sy(2 downto 0) = "000"))) else '0';

end arch_controlcursor;
