--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;
use work.fpoinc_pkg.all;
use work.fpocfg_pkg.all;

package graphics_pkg is

  component graphics is
    port (vclk, vena, ready, fill, run : in std_ulogic; badr : in ramadr_typ; vadr : out ramadr_typ; d : fpo_data_typ; capture_ctrl : in capture_ctrl_rec; display_ctrl : in display_ctrl_rec; drift : drift_typ; xcursor : in col_typ; vc : out vgaout; vact, vpass : out std_ulogic);
  end component;

end graphics_pkg;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;
use work.videomode_pkg.all;
use work.rastersync_pkg.all;
use work.videoadr_pkg.all;
use work.gobjs_pkg.all;
use work.main_pkg.all;
use work.tracescan_pkg.all;
use work.tracedmux_pkg.all;
use work.chargen_pkg.all;
use work.traceval_pkg.all;

entity graphics is
  port (vclk, vena, ready, fill, run : in std_ulogic; badr : in ramadr_typ; vadr : out ramadr_typ; d : fpo_data_typ; capture_ctrl : in capture_ctrl_rec; display_ctrl : in display_ctrl_rec; drift : drift_typ; xcursor : in col_typ; vc : out vgaout; vact, vpass : out std_ulogic);
end graphics;

architecture arch_graphics of graphics is

  constant vbuf : boolean := true;

  constant hemthick : positive := 4;

  signal vactive, hactive, next_line, gl_blank, pass,
    text_strobe, gridpix, charpix, trgpix, ctrlcpix, timecursor,
    tvr_start, tvc_start, tvr_strobe, markdrift, driftstrobe,
    mstrobe, tstrobe, hemstrobe, valback, next_string, rightval,
    rowstart, statrowstart, colstart, statcolstart, modecolstart, trgcolstart,
    last_char, mode_last_char, trg_last_char, stat_last_char, hex_last_char : std_ulogic;
  signal sel_valid                                 : boolean;
  signal rpos, rpos1, timecursorpos                : ramadr_typ;
  signal col                                       : col_typ;
  signal row                                       : row_typ;
  signal stat_chr, trg_chr, mode_chr, hex_chr, chr : character;
  signal mcolor                                    : color_bus_typ;
  signal trace_count                               : trace_range_typ;
  signal line                                      : std_ulogic_vector(7 downto 0);
  signal selbus                                    : selbus_typ;
  signal tracedbus                                 : tracedbus_typ;
  signal linenum                                   : linenum_typ;
  signal colnum                                    : colnum_typ;
  signal charnum                                   : charnum_typ;
  signal mode_color, trg_color                     : color_typ;
  signal vint                                      : vgaout;
  signal timecursordata                            : fpo_data_typ;

begin

  rastersync_ins : rastersync port map(
    clk       => vclk,
    ena       => vena,
    hactive   => hactive,
    vactive   => vactive,
    next_line => next_line,
    hsync     => vint.h,
    vsync     => vint.v,
    vcol      => col,
    vrow      => row);

  videoadr_ins : videoadr port map(
    clk          => vclk,
    ena          => vena,
    pass         => pass,
    hactive      => hactive,
    vactive      => vactive,
    capture_ctrl => capture_ctrl,
    display_ctrl => display_ctrl,
    badr         => badr,
    blank        => gl_blank,
    adr          => vadr,
    count        => rpos);

  process(vclk)
  begin
    if rising_edge(vclk) then
      if '1' = vena then
        if '1' = pass then
          rpos1 <= rpos;
        end if;
      end if;
    end if;
  end process;

  process(col, display_ctrl.scale, hactive)
  begin
    if '1' /= hactive then
      pass <= '1';
    else
      case display_ctrl.scale is
        when 0      => pass <= '1';
        when 1      => pass <= col(0);
        when 2      => pass <= col(0) and col(1);
        when others => pass <= '1';
      end case;
    end if;
  end process;

  vpass <= pass;

  tracescan_ins : tracescan port map(
    clk       => vclk,
    ena       => vena,
    vactive   => vactive,
    next_line => next_line,
    line      => line,
    trace     => trace_count,
    valid     => sel_valid);

  tracedmux_ins : tracedmux port map(
    i         => trace_count,
    e         => sel_valid,
    d         => d,
    selbus    => selbus,
    tracedbus => tracedbus);

  main_ins : main port map(
    clk       => vclk,
    ena       => vena,
    line      => line,
    selbus    => selbus,
    tracedbus => tracedbus,
    scale     => display_ctrl.scale,
    col       => col,
    color     => mcolor);

  gridline_ins : gridline port map(
    clk  => vclk,
    ena  => vena,
    pix  => gridpix,
    col  => col,
    line => line(3 downto 0));

  tvg : if traceval_enable generate
    tracevalrow_ins : tracevalrow port map(
      i     => trace_count,
      line  => line,
      start => tvr_start);

    tracevalcol_ins : tracevalcol port map(
      rightval => rightval,
      col      => col,
      start    => tvc_start);

    rowff_ins : row_ff port map(
      clk         => vclk,
      ena         => vena,
      next_line   => next_line,
      rowstart    => tvr_start,
      next_string => next_string,
      last_string => '1',
      colstart    => '0',               --tvc_start,
      subcolstart => open,
      rowstrobe   => tvr_strobe);
  end generate;

  hexstring_ins : hexstring port map(
    charnum   => charnum,
    i         => trace_count,
    d         => timecursordata,
    last_char => hex_last_char,
    chr       => hex_chr);

  statusstring_ins : statusstring port map(
    charnum      => charnum,
    capture_ctrl => capture_ctrl,
    display_ctrl => display_ctrl,
    cur          => timecursorpos,
    last_char    => stat_last_char,
    chr          => stat_chr);

  trgstring_ins : trgstring port map(
    clk           => vclk,
    ena           => vena,
    vactive       => vactive,
    ready         => ready,
    fill          => fill,
    run           => run,
    charnum       => charnum,
    last_char     => trg_last_char,
    chr           => trg_chr,
    trigger_color => trg_color);

  modestring_ins : modestring port map(
    trg_mod    => capture_ctrl.mode,
    charnum    => charnum,
    last_char  => mode_last_char,
    chr        => mode_chr,
    mode_color => mode_color);

  cs_m : colstrobe port map(
    clk      => vclk,
    ena      => vena,
    col      => col,
    colbegin => std_ulogic_vector(to_unsigned(videotiming.hvisible - (fontcfg.width * 8), col_typ'length)),
    colend   => std_ulogic_vector(to_unsigned(videotiming.hvisible - (fontcfg.width * 4) - 1, col_typ'length)),
    strobe   => mstrobe);

  cs_t : colstrobe port map(
    clk      => vclk,
    ena      => vena,
    col      => col,
    colbegin => std_ulogic_vector(to_unsigned(videotiming.hvisible - (fontcfg.width * 3), col_typ'length)),
    colend   => std_ulogic_vector(to_unsigned(videotiming.hvisible - 1, col_typ'length)),
    strobe   => tstrobe);

  rs : rowstrobe port map(
    clk       => vclk,
    ena       => vena,
    next_line => next_line,
    row       => row,
    rowbegin  => std_ulogic_vector(to_unsigned(videotiming.vvisible - (hemthick * 2 + fontcfg.height), col_typ'length)),
    rowend    => std_ulogic_vector(to_unsigned(videotiming.vvisible - 1, col_typ'length)),
    strobe    => hemstrobe);

  statrowstart <= '1' when ((videotiming.vvisible - fontcfg.height - hemthick) = to_integer(unsigned(row))) else '0';
  statcolstart <= '1' when 0 = to_integer(unsigned(col))                                                    else '0';
  modecolstart <= '1' when ((videotiming.hvisible - (fontcfg.width * 8)) = to_integer(unsigned(col)))       else '0';
  trgcolstart  <= '1' when ((videotiming.hvisible - (fontcfg.width * 3)) = to_integer(unsigned(col)))       else '0';

  rowstart <= statrowstart or tvr_start;

  colstart <= tvc_start when '1' = tvr_strobe else (statcolstart or modecolstart or trgcolstart);

  valback <= text_strobe and tvr_strobe;

  chr <= mode_chr when ('1' = mstrobe) else trg_chr when ('1' = tstrobe) else hex_chr when ('1' = tvr_strobe) else stat_chr;

  last_char <= (mode_last_char and mstrobe) or (trg_last_char and tstrobe) or stat_last_char or (hex_last_char and tvr_strobe);

  stringgenerator_ins : string_generator port map(
    clk         => vclk,
    ena         => vena,
    next_line   => next_line,
    rowstart    => rowstart,
    colstart    => colstart,
    last_char   => last_char,
    last_string => '1',
    linenum     => linenum,
    colnum      => colnum,
    charnum     => charnum,
    next_string => next_string,
    strobe      => text_strobe);

  controlcursor_ins : controlcursor port map(
    clk => vclk,
    ena => vena,
    pix => ctrlcpix,
    x   => xcursor,
    col => col,
    y   => std_ulogic_vector(to_unsigned(videotiming.vvisible - (fontcfg.height / 2 + hemthick), row'length)),
    row => row);

  timecursor <= '1' when xcursor = col else '0';

  process(vclk)
  begin
    if rising_edge(vclk) then
      if '1' = vena then
        if '1' = timecursor then
          timecursordata <= d;
          timecursorpos  <= rpos1;
        end if;
      end if;
    end if;
  end process;

  rightval <= '1' when (videotiming.hvisible / 2) > to_integer(unsigned(xcursor)) else '0';

  trgpix <= '1' when drift = rpos1 else '0';

  markdrift <= '1' when drift /= capture_ctrl.drift else '0';

  cs_d : colstrobe port map(
    clk      => vclk,
    ena      => vena,
    col      => col,
    colbegin => std_ulogic_vector(to_unsigned((fontcfg.width * 16), col_typ'length)),
    colend   => std_ulogic_vector(to_unsigned((fontcfg.width * (16 + 11) - 1), col_typ'length)),
    strobe   => driftstrobe);

  charactergenerator_ins : character_generator port map(
    chr     => chr,
    linenum => linenum,
    colnum  => colnum,
    pix     => charpix);

  vint.r <= ((charpix and valback) or ((not valback) and (ctrlcpix or (((((mcolor.r and not gl_blank) xor gridpix xor trgpix) or timecursor) and mcolor.pix) xor (charpix and text_strobe and not (mstrobe or tstrobe))                                       xor (not (charpix and text_strobe) and ((mode_color.r and mstrobe) or (trg_color.r and tstrobe)) and hemstrobe))))) and hactive and vactive;
  vint.g <= ((charpix and valback) or ((not valback) and (            ((( (mcolor.g and not gl_blank) xor             trgpix               ) and mcolor.pix) xor (charpix and text_strobe and not (mstrobe or tstrobe) and (not (markdrift and driftstrobe))) xor (not (charpix and text_strobe) and ((mode_color.g and mstrobe) or (trg_color.g and tstrobe)) and hemstrobe))))) and hactive and vactive;
  vint.b <= (                         ((not valback) and (            ((( (mcolor.b and not gl_blank) xor gridpix                          ) and mcolor.pix)                                                                                                  xor (not (charpix and text_strobe) and ((mode_color.b and mstrobe) or (trg_color.b and tstrobe)) and hemstrobe))))) and hactive and vactive;

  vact <= vactive;

  vb : if vbuf generate
    process(vclk)
    begin
      if rising_edge(vclk) then
        if '1' = vena then
          vc <= vint;
        end if;
      end if;
    end process;
  end generate;

  nvb : if not vbuf generate
    vc <= vint;
  end generate;

end arch_graphics;
