--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;
use work.fpoinc_pkg.all;

package mousectrl_pkg is

  component mousectrl is
    port (clk, ena, reset : in std_ulogic; ps2c, ps2d : inout std_ulogic; capture_ctrl : out capture_ctrl_rec; display_ctrl : out display_ctrl_rec; xcursor : out col_typ; update : out std_ulogic);  --update
  end component;

end mousectrl_pkg;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.videomode_pkg.all;
use work.fpotyp_pkg.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;
use work.ps2mouse_pkg.all;

entity mousectrl is
  port (clk, ena, reset : in std_ulogic; ps2c, ps2d : inout std_ulogic; capture_ctrl : out capture_ctrl_rec; display_ctrl : out display_ctrl_rec; xcursor : out col_typ; update : out std_ulogic);  --update
end mousectrl;

architecture arch_mousectrl of mousectrl is

  constant scale_left  : positive := 8 * 1;
  constant scale_right : positive := 8 * 13;

  constant drift_left  : positive := 8 * 16;
  constant drift_right : positive := 8 * 27;

  constant disp_left  : positive := 8 * 30;
  constant disp_right : positive := 8 * 38;

  constant mode_left  : positive := (videotiming.hvisible/32 - 2) * 32 + 8;
  constant mode_right : positive := mode_left + 16;

  constant act_left  : positive := (videotiming.hvisible/32 - 1) * 32 + 16;
  constant act_right : positive := act_left + 16;

  constant xmax : unsigned(wh(videotiming.hvisible) downto 0) := to_unsigned(videotiming.hvisible - 1, col_typ'length);

  signal x                                                                : unsigned(wh(videotiming.hvisible) downto 0);
  signal movedata                                                         : movedata_rec;
  signal mousevalid, mousevalid_ff                                        : std_ulogic;
  signal btn_left_s, btn_left_ff, btn_left_click                          : boolean;
  signal scale_select, drift_select, disp_select, mode_select, act_select : boolean;
  signal scale                                                            : scale_typ                                                                           := scale_default;
  signal drift                                                            : drift_typ                                                                           := drift_default;
  signal disp                                                             : ramadr_typ                                                                          := disp_default;
--  signal trgmode : trg_mod_typ := trgmode_default;
  signal act                                                              : std_ulogic                                                                          := act_default;
  signal depth                                                            : ramadr_typ                                                                          := depth_default;
  signal trgmode_pos                                                      : natural range trg_mod_typ'pos(trg_mod_typ'low) to trg_mod_typ'pos(trg_mod_typ'high) := trg_mod_typ'pos(trgmode_default);

begin

  mouse_move : mousemove port map(
    clk      => clk,
    ena      => ena,
    reset    => reset,
    ps2c     => ps2c,
    ps2d     => ps2d,
    movedata => movedata,
    valid    => mousevalid);

  process(clk, reset)
  begin
    if '1' = reset then
      mousevalid_ff <= '0';
    else
      if rising_edge(clk) then
        if '1' = ena then
          mousevalid_ff <= mousevalid;
        end if;
      end if;
    end if;
  end process;

  btn_left_s <= '1' = movedata.btns(btn_left);

  process(clk, reset)
  begin
    if '1' = reset then
      btn_left_ff <= false;
    else
      if rising_edge(clk) then
        if '1' = ena then
          if '1' = (mousevalid and not mousevalid_ff) then
            btn_left_ff <= btn_left_s;
          end if;
        end if;
      end if;
    end if;
  end process;

  btn_left_click <= btn_left_s and not btn_left_ff;

  scale_select <= (scale_left <= x) and (scale_right >= x);
  drift_select <= (drift_left <= x) and (drift_right >= x);
  disp_select  <= (disp_left <= x) and (disp_right >= x);
  mode_select  <= (mode_left <= x) and (mode_right >= x);
  act_select   <= (act_left <= x) and (act_right >= x);

  process(clk)
    variable xt : unsigned(wh(videotiming.hvisible) downto 0);
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if '1' = (mousevalid and not mousevalid_ff) then
          if ('0' = movedata.btns(btn_left)) then
            xt := unsigned(signed(x) + movedata.x);
            if to_signed(0, movedata.x'length) > movedata.x then
              if (x < xt) or (xmax < xt) then
                x <= (others => '0');
              else
                x <= xt;
              end if;
            else
              if (x > xt) or (xmax < xt) then
                x <= xmax;
              else
                x <= xt;
              end if;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;


  process(clk, reset)
    variable xt     : unsigned(wh(videotiming.hvisible) downto 0);
    variable driftt : drift_typ;
    variable dispt  : ramadr_typ;
  begin
    if '1' = reset then

      scale       <= scale_default;
      drift       <= drift_default;
      disp        <= disp_default;
--      trgmode <= trgmode_default;
      trgmode_pos <= trg_mod_typ'pos(trgmode_default);
      act         <= act_default;
      depth       <= depth_default;

    else

      if rising_edge(clk) then
        if '1' = ena then
          if '1' = (mousevalid and not mousevalid_ff) then

            if scale_select and btn_left_click then
              if scale_typ_high <= scale then
                scale <= scale_typ_low;
              else
                scale <= scale + 1;
              end if;
            end if;

            if drift_select and btn_left_s then
              driftt := drift + to_integer(movedata.y);
              if to_signed(0, movedata.y'length) > movedata.y then
                if (drift < driftt) or (drift_typ_low > driftt) then  -- depth
                  drift <= drift_typ_low;
                else
                  drift <= driftt;
                end if;
              else
                if (drift > driftt) or (drift_typ_high < driftt) then
                  drift <= drift_typ_high;
                else
                  drift <= driftt;
                end if;
              end if;
            end if;

            if disp_select and btn_left_s then
              dispt := disp + to_integer(movedata.y);
              if to_signed(0, movedata.y'length) > movedata.y then
                if (disp < dispt) or (ramadr_typ_low > dispt) then
                  disp <= ramadr_typ_low;
                else
                  disp <= dispt;
                end if;
              else
                if (disp > dispt) or (depth - 100 < dispt) then
                  disp <= depth - 100;
                else
                  disp <= dispt;
                end if;
              end if;
            end if;

            if mode_select and btn_left_click then
              if trg_mod_typ'pos(trg_mod_typ'high) <= trgmode_pos then
                trgmode_pos <= trg_mod_typ'pos(trg_mod_typ'low);
              else
                trgmode_pos <= trgmode_pos + 1;
              end if;
            end if;

            if act_select and btn_left_click then
              act <= not act;
            end if;

          end if;
        end if;
      end if;
    end if;
  end process;

  display_ctrl.scale <= scale;

  capture_ctrl.drift <= drift;

  display_ctrl.disp <= disp;

  capture_ctrl.mode <= sngl when (1 = trgmode_pos) else norm;  --trg_mod_typ'val(trgmode_pos); --trgmode;

  capture_ctrl.act <= act;

  capture_ctrl.depth <= depth;

  xcursor <= std_ulogic_vector(x);

  update <= mousevalid;

end arch_mousectrl;
