--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use work.videomode_pkg.all;
use work.fpoinc_pkg.all;
use work.chargen_pkg.all;

package traceval_pkg is

  component tracevalrow is
    port(i : in trace_range_typ; line : in std_ulogic_vector(7 downto 0); start : out std_ulogic);
  end component;

  component tracevalcol is
    port(rightval : in std_ulogic; col : in col_typ; start : out std_ulogic);
  end component;

  component hexstring is
    port (charnum : in charnum_typ; i : in trace_range_typ; d : in fpo_data_typ; last_char : out std_ulogic; chr : out character);
  end component;

end traceval_pkg;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.chargen_pkg.all;
use work.fpotyp_pkg.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;

entity tracevalrow is
  port(i : in trace_range_typ; line : in std_ulogic_vector(7 downto 0); start : out std_ulogic);
end tracevalrow;

architecture arch_tracevalrow of tracevalrow is

  type trace_valline_typ is array (trace_typ_typ) of natural;

  constant trace_valline : trace_valline_typ := (
    tt_bit16   => (trace_properties(tt_bit16).lines - fontcfg.height) / 2,
    tt_bus16w4 => (trace_properties(tt_bus16w4).lines - fontcfg.height) / 2,
    tt_dot64u  => (trace_properties(tt_dot64u).lines - fontcfg.height) / 2,
    tt_bar64u  => (trace_properties(tt_bar64u).lines - fontcfg.height) / 2,
    tt_bar256u => (trace_properties(tt_bar256u).lines - fontcfg.height) / 2,
    tt_separ4  => 4);

begin

  start <= '1' when std_ulogic_vector(to_unsigned(trace_valline(fpo_item_typ(i)), 8)) = line else '0';

end arch_tracevalrow;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.videomode_pkg.all;
use work.chargen_pkg.all;

entity tracevalcol is
  port(rightval : in std_ulogic; col : in col_typ; start : out std_ulogic);
end tracevalcol;

architecture arch_tracevalcol of tracevalcol is

  signal l, r : std_ulogic;

begin

  l <= '1' when (videotiming.hvisible / 4 - fontcfg.width) = to_integer(unsigned(col))                                else '0';
  r <= '1' when ((videotiming.hvisible / 2) + (videotiming.hvisible / 4) - fontcfg.width) = to_integer(unsigned(col)) else '0';

  start <= r when '1' = rightval else l;

end arch_tracevalcol;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.chargen_pkg.all;
use work.fpotyp_pkg.all;
use work.fpocfg_pkg.all;
use work.fpoinc_pkg.all;

entity hexstring is
  port (charnum : in charnum_typ; i : in trace_range_typ; d : in fpo_data_typ; last_char : out std_ulogic; chr : out character);
end hexstring;

architecture arch_hexstring of hexstring is

  subtype nibble_typ is std_ulogic_vector(3 downto 0);

  subtype hexleng_typ is natural range 0 to 1;

  type nibble_arr is array (hexleng_typ) of nibble_typ;

  type char_map_typ is array (0 to 15) of character;
  constant char_map : char_map_typ := ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');

  function nibbleng(w : natural) return natural is
    variable l : natural;
  begin
    l := w / nibble_typ'length;
    if l * nibble_typ'length < w then
      l := l + 1;
    end if;
    return l;
  end;

  type trace_nlast_typ is array (trace_typ_typ) of natural;

  constant trace_nlast : trace_nlast_typ := (
    tt_bit16   => nibbleng(trace_properties(tt_bit16).width) - 1,
    tt_bus16w4 => nibbleng(trace_properties(tt_bus16w4).width) - 1,
    tt_dot64u  => nibbleng(trace_properties(tt_dot64u).width) - 1,
    tt_bar64u  => nibbleng(trace_properties(tt_bar64u).width) - 1,
    tt_bar256u => nibbleng(trace_properties(tt_bar256u).width) - 1,
    tt_separ4  => 0);  --nibbleng(trace_properties(tt_separ4).width) - 1);

  function nibbling2(d : std_ulogic_vector) return nibble_arr is
    variable x : std_ulogic_vector(d'length - 1 downto 0) := d;
    variable k : integer;
    variable n : nibble_arr;
    variable l : natural := nibbleng(d'length);
  begin
    for i in nibble_arr'range loop
      for j in nibble_typ'range loop
        k := (nibble_typ'length * (l - i)) - (nibble_typ'length - j);
        if (d'length > k) and (0 <= k) then
          n(i)(j) := x(k);
        else
          n(i)(j) := '0';
        end if;
      end loop;
    end loop;
    return n;
  end;

  function nmux(i : trace_range_typ; d : fpo_data_typ) return nibble_arr is

    subtype fpo_data_range_typ is natural range fpo_data_typ'range;

    type fpo_item_high_typ is array (trace_range_typ) of fpo_data_range_typ;

    type nibble_arr_arr is array (trace_range_typ) of nibble_arr;

    variable low, fpo_item_high : fpo_item_high_typ;
    variable n                  : nibble_arr_arr;
    variable l                  : natural;
    variable m                  : integer;

  begin
    for j in trace_range_typ loop
      fpo_item_high(j) := fpo_item_low(j) + fpo_item_width(j) - 1;
      low(j)           := fpo_item_low(j);
    end loop;

    for j in trace_range_typ loop
      l := nibbleng(fpo_item_width(j));
      for k in nibble_arr'range loop
        for t in nibble_typ'range loop
          m                                 := ((l - k - 1) * nibble_typ'length) + t;
          if (fpo_item_width(j) > m) and (0 <= m) then
            n(j)(k)(t) := d(low(j) + m);
          else
            n(j)(k)(t) := '0';
          end if;
        end loop;
      end loop;
    end loop;
    return n(i);
  end;

begin

  chr <= char_map(to_integer(unsigned(nmux(i, d)(charnum rem 2))));

  last_char <= '1' when trace_nlast(fpo_item_typ(i)) = (charnum rem 2) else '0';

end arch_hexstring;
