--Copyright (C) 2015 Zen Lu and released under the GPLv3

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package ps2mouse_pkg is

  type btn_enum is (btn_left, btn_right, btn_middle);

  type btn_array is array (btn_enum) of std_ulogic;

  type movedata_rec is record
    x, y : signed(8 downto 0);
    btns : btn_array;
  end record;

  component mousemove is
    port (clk, ena, reset : in std_ulogic; ps2c, ps2d : inout std_ulogic; movedata : out movedata_rec; valid : out std_ulogic);
  end component;

end ps2mouse_pkg;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.ps2mouse_pkg.all;
use work.ps2host_pkg.all;

entity mousemove is
  port (clk, ena, reset : in std_ulogic; ps2c, ps2d : inout std_ulogic; movedata : out movedata_rec; valid : out std_ulogic);
end mousemove;

architecture arch_mousemove of mousemove is

  constant divisor   : positive := 2 ** 17 - 1;
  constant timer_max : positive := 255;

  signal divider                                        : natural range 0 to divisor - 1;
  signal d, s_reg, x_reg, y_reg                         : std_ulogic_vector(7 downto 0);
  signal strobe, txbusy, valid_rx, valid_ff0, valid_ff1 : std_ulogic;
  signal valid_strobe, timer_stop, load0, step          : boolean;
  signal timer                                          : natural range 0 to timer_max;
  signal state                                          : natural range 0 to 3;

begin

  process(clk, reset)
  begin
    if '1' = reset then
      divider <= 0;
    else
      if rising_edge(clk) then
        if '1' = ena then
          if strobe = '1' then
            divider <= 0;
          else
            divider <= divider + 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  strobe <= '1' when (divisor - 1) = divider else '0';

  process(clk, reset)
  begin
    if '1' = reset then
      valid_ff0 <= '0';
      valid_ff1 <= '0';
    else
      if rising_edge(clk) then
        if '1' = ena then
          valid_ff0 <= valid_rx;
          valid_ff1 <= valid_ff0;
        end if;
      end if;
    end if;
  end process;

  valid_strobe <= '1' = (valid_ff0 and not valid_ff1);

  process(clk, reset)
  begin
    if '1' = reset then
      timer <= 0;
    else
      if rising_edge(clk) then
        if '1' = ena then
          if valid_strobe then
            timer <= 0;
          else
            if not timer_stop then
              timer <= timer + 1;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  timer_stop <= (timer_max = timer);

  load0 <= timer_stop when 0 /= state else false;

  process(clk, reset)
  begin
    if '1' = reset then
      state <= 0;
    else
      if rising_edge(clk) then
        if '1' = ena then
          if load0 then
            state <= 0;
          else
            if step then
              state <= state + 1;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  process(state, valid_strobe, timer_stop)
  begin
    case state is
      when 0 =>
        step <= timer_stop and valid_strobe;
      when 1 | 2 =>
        step <= valid_strobe;
      when 3 =>
        step <= false;
    end case;
  end process;

  process(clk)
  begin
    if rising_edge(clk) then
      if '1' = ena then
        if step then
          case state is
            when 0 =>
              s_reg <= d;
            when 1 =>
              x_reg <= d;
            when 2 =>
              y_reg <= d;
            when 3 =>
              null;
          end case;
        end if;
      end if;
    end if;
  end process;

  b : if true generate
    process(clk, reset)
    begin
      if '1' = reset then
        valid <= '0';
      else
        if rising_edge(clk) then
          if '1' = ena then
            if step then
              case state is
                when 0 =>
                  valid <= '0';
                when 2 =>
                  valid <= '1';
                when others =>
                  null;
              end case;
            end if;
          end if;
        end if;
      end if;
    end process;
  end generate;

  rx : ps2hrnc port map(
    reset => txbusy,
    ps2c  => ps2c,
    ps2d  => ps2d,
    d     => d,
    v     => valid_rx);

  tx : ps2_htx port map(
    clk    => clk,
    ena    => ena,
    reset  => reset,
    strobe => strobe,
    ps2c   => ps2c,
    ps2d   => ps2d,
    d      => X"F4", --enable streaming,
    ack    => open,
    busy   => txbusy);

  movedata.x                <= signed(s_reg(4) & x_reg);
  movedata.y                <= signed(s_reg(5) & y_reg);
  movedata.btns(btn_left)   <= s_reg(0);
  movedata.btns(btn_right)  <= s_reg(1);
  movedata.btns(btn_middle) <= s_reg(2);

end arch_mousemove;
